function [uv canInd] = combine_flow_proposals(this, puv, tu, tv, niters)
%COMBINE_FLOW_PROPOSALS combines the flow proposals using the combined
%local minima (CLM) method by Jung et al. ECCV2008

% Abandon outliers in proposals?

if nargin <5
    niters       = 100;      % iteration number
end;

maxW         = round(linspace(40, 40, niters));       % maximum window size for combination
minW         = round(linspace(1,   1, niters));       % minimum window size for combination  

this.display = false;
isz         = [size(puv{1}, 1) size(puv{2},2)];

% Precompute 
nP = length(puv);

% Data term
data = cell(1,nP);
for i = 1:nP
    It      = partial_deriv(this.images, puv{i}, this.interpolation_method, this.deriv_filter);
    temp    = reshape(evaluate_log(this.rho_data, It(:)'), size(It));    
    % pay a constant penalty for those mapped out of boundaries to
    % discourage too many out-of-boundary pixels 
    indx    = It == 0;
    pen     = mean(temp(~indx));
    temp(indx) = pen;
    data{i} = temp;
end;

% Spatial term
S = this.spatial_filters;       %{[1 -1], [1; -1]};

spatialx = cell(1,nP);
spatialy = spatialx;
for i = 1:nP
    ux = conv2(puv{i}(:,:,1), S{1}, 'same');
    vx = conv2(puv{i}(:,:,2), S{1}, 'same');
    uy = conv2(puv{i}(:,:,1), S{2}, 'same');
    vy = conv2(puv{i}(:,:,2), S{2}, 'same');
   
    sux = reshape(evaluate_log(this.rho_spatial_u{1}, ux(:)'), size(ux));
    svx = reshape(evaluate_log(this.rho_spatial_v{1}, vx(:)'), size(ux));
    suy = reshape(evaluate_log(this.rho_spatial_u{2}, uy(:)'), size(ux));
    svy = reshape(evaluate_log(this.rho_spatial_v{2}, vy(:)'), size(ux));

    % Remove these invalid regions
    sux(:,end) = 0;
    suy(end,:) = 0;
    svx(:,end) = 0;
    svy(end,:) = 0;

    spatialx{i} = sux+svx;
    spatialy{i} = suy+svy;
end;

% start from the one with high log posterior
logp = zeros(1,nP);
for i = 1:nP
    logp(i) = sum(data{i}(:)) + this.lambda*sum(spatialx{i}(:)+spatialy{i}(:));
end;
[tmp ind] = max(logp);
% [tmp ind] = min(logp);             % start from lowest posteriori
% ind = max(round(nP*rand()), 1);         % start from random initialization

uv = puv{ind};
uv_data = data{ind};
uv_spax = spatialx{ind};
uv_spay = spatialy{ind};

fprintf('Begin CLM: log-posterior\t%3.2e\n', evaluate_log_posterior(this, uv));
if nargin >= 4
    [aae stdae aepe] = flowAngErr(tu, tv, uv(:,:,1), uv(:,:,2), 0);
    fprintf('AAE %3.3f STD %3.3f average end point error %3.3f \n\n', aae, stdae, aepe);
end;

% CLM

canInd  = ind*ones(isz);
  
logP    = zeros(1,nP+1);
for itr = 1:niters;

    % Generate random window paprtition    
%     psz(1) = max(round(rand*maxW), 1);      % partition window size
%     psz(2) = max(round(rand*maxW), 1);
    psz(1) = max(round(rand*maxW(itr)), minW(itr));
    psz(2) = max(round(rand*maxW(itr)), minW(itr));
    
    sht(1) = max(round(rand*psz(1)), 1);    % partition window shfit
    sht(2) = max(round(rand*psz(2)), 1);
   
    npr  = ceil( (isz(1)-sht(1)) / psz(1) )+1;
    npc  = ceil( (isz(2)-sht(2)) / psz(2) )+1;
    
    for i = 1:npr
        for j = 1:npc
            
            x1 = max( (j-2)*psz(2)+1+sht(2), 1 );
            x2 = min( (j-1)*psz(2)+sht(2)  , isz(2));
            y1 = max( (i-2)*psz(1)+1+sht(1), 1);
            y2 = min( (i-1)*psz(1)+sht(1)  , isz(1));
            
            % log posterior (negative energy) of the proposal partition
            for k = 1:nP
                ed = sum(sum(data{k}(y1:y2, x1:x2)));
                es = sum(sum(spatialx{k}(y1:y2, x1:x2))) + sum(sum(spatialy{k}(y1:y2, x1:x2)))... % not use y2-1 to avoid situation y1 == y2
                     -   sum(spatialx{k}(y1:y2, x2))     -     sum(spatialy{k}(y2, x1:x2));
                 %...                     +       spatialy{k}(y2   , x2); 
                
                % Interaction with those pixels surrounding the current patch
                
                % evalaute_log function of GSM takes in an p by ndata
                % vector and returns a 1 by n vector
                if x1 > 1
                    es = es + sum(evaluate_log(this.rho_spatial_u{1}, puv{k}(y1:y2, x1, 1)' - uv(y1:y2, x1-1, 1)'));
                    es = es + sum(evaluate_log(this.rho_spatial_v{1}, puv{k}(y1:y2, x1, 2)' - uv(y1:y2, x1-1, 2)'));                    
                end;
                if x2 < isz(2)
                    es = es + sum(evaluate_log(this.rho_spatial_u{1}, puv{k}(y1:y2, x2, 1)' - uv(y1:y2, x2+1, 1)'));
                    es = es + sum(evaluate_log(this.rho_spatial_v{1}, puv{k}(y1:y2, x2, 2)' - uv(y1:y2, x2+1, 2)')); 
                end;                
                
                if y1 > 1
                    es = es + sum(evaluate_log(this.rho_spatial_u{2}, puv{k}(y1, x1:x2, 1) - uv(y1-1, x1:x2, 1)));
                    es = es + sum(evaluate_log(this.rho_spatial_v{2}, puv{k}(y1, x1:x2, 2) - uv(y1-1, x1:x2, 2)));
                end;
                
                if y2 < isz(1)
                    es = es + sum(evaluate_log(this.rho_spatial_u{2}, puv{k}(y2, x1:x2, 1) - uv(y2+1, x1:x2, 1)));
                    es = es + sum(evaluate_log(this.rho_spatial_v{2}, puv{k}(y2, x1:x2, 2) - uv(y2+1, x1:x2, 2)));
                end;
                
                logP(k) = ed + this.lambda*es;
            end;            
            
            % log posterior of the current window of the current solution
            
            xt = max(x1-1, 1);
            yt = max(y1-1, 1);           

            es = sum(sum(uv_spax(y1:y2, xt:x2))) + sum(sum(uv_spay(yt:y2, x1:x2)));           
            ed = sum(sum(uv_data(y1:y2, x1:x2)));            
            logP(nP+1) = ed + this.lambda*es;            
            
            [tmp ind] = max(logP); 

            if ind ~= nP+1
                canInd(y1:y2, x1:x2,:)      = ind;  % record where the candinate comes from
                
                % update
                uv(y1:y2, x1:x2,:)          = puv{ind}(y1:y2, x1:x2,:);
                uv_data(y1:y2, x1:x2)       = data{ind}(y1:y2, x1:x2);
                uv_spax(y1:y2, x1:x2)       = spatialx{ind}(y1:y2, x1:x2);
                uv_spay(y1:y2, x1:x2)       = spatialy{ind}(y1:y2, x1:x2);
                
                % evalaute_log function of GSM takes in an p by ndata
                % vector and returns a 1 by n vector
                
                if x1 > 1
                    uv_spax(y1:y2, x1-1)    = evaluate_log(this.rho_spatial_u{1}, uv(y1:y2, x1, 1)' - uv(y1:y2, x1-1, 1)') +...
                                              evaluate_log(this.rho_spatial_v{1}, uv(y1:y2, x1, 2)' - uv(y1:y2, x1-1, 2)');
                end;
                if x2 < isz(2)
                    uv_spax(y1:y2, x2)      = evaluate_log(this.rho_spatial_u{1}, uv(y1:y2, x2, 1)' - uv(y1:y2, x2+1, 1)') + ...;
                                              evaluate_log(this.rho_spatial_v{1}, uv(y1:y2, x2, 2)' - uv(y1:y2, x2+1, 2)'); 
                end;
                if y1 > 1
                    uv_spay(y1-1, x1:x2)    = evaluate_log(this.rho_spatial_u{2}, uv(y1, x1:x2, 1) - uv(y1-1, x1:x2, 1)) +...
                                              evaluate_log(this.rho_spatial_v{2}, uv(y1, x1:x2, 2) - uv(y1-1, x1:x2, 2));
                end;
                if y2 < isz(1)
                    uv_spay(y2, x1:x2)      = evaluate_log(this.rho_spatial_u{2}, uv(y2, x1:x2, 1) - uv(y2+1, x1:x2, 1)) +...
                                              evaluate_log(this.rho_spatial_v{2}, uv(y2, x1:x2, 2) - uv(y2+1, x1:x2, 2));
                end;
                
            end;
            
%             % For GSM models these invalid pixels are also updated?
%             uv_spax(:,end) = 0;
%             uv_spay(end,:) = 0;

        end;
    end;
    
    fprintf('iter --- %d window size --- %d, %d \tlog-posterior\t%3.2e\n', itr, psz(1), psz(2), evaluate_log_posterior(this, uv));
    if nargin >= 4

        [aae stdae aepe] = flowAngErr(tu, tv, uv(:,:,1), uv(:,:,2), 0);
        fprintf('AAE %3.3f STD %3.3f average end point error %3.3f \n\n', aae, stdae, aepe);
    end;
end;


% To do: replace with conjugate gradient and line search
% fprintf('Start performing gradient ascent \n');
% nGDiters = 100;
% tau    = 5e-3;
% for i = 1:nGDiters
%     duv = evaluate_log_posterior_grad(this, uv);
%     uv  = uv + duv*tau;
%     if mod(i, nGDiters/10) == 0
%         fprintf('log-posterior\t%3.3e\n', evaluate_log_posterior(ba, uv));
%         if nargin == 4
%             [aae stdae aepe] = flowAngErr(tu, tv, uv(:,:,1), uv(:,:,2), 0);
%             fprintf('AAE %3.3f STD %3.3f average end point error %3.3f \n\n', aae, stdae, aepe);
%         end;
%     end;
% end;