function [uv seg]= compute_flow(this, init, gt)
% function output= compute_flow(this, init, gt)
%%
%COMPUTE_FLOW   Compute flow field
%   UV = COMPUTE_FLOW(THIS[, INIT]) computes the flow field UV with
%   algorithm THIS and the optional initialization INIT.
%  
%   This is a member function of the class 'layer_optical_flow'. 

%   Author: Deqing Sun, Department of Computer Science, Brown University
%   Contact: dqsun@cs.brown.edu
%   $Date: 2007-- $

% Copyright 2007-2008, Brown University, Providence, RI.
%
%                         All Rights Reserved
%
% Permission to use, copy, modify, and distribute this software and its
% documentation for any purpose other than its incorporation into a
% commercial product is hereby granted without fee, provided that the
% above copyright notice appear in all copies and that both that
% copyright notice and this permission notice appear in supporting
% documentation, and that the name of the author and Brown University not be used in
% advertising or publicity pertaining to distribution of the software
% without specific, written prior permission.
%
% THE AUTHOR AND BROWN UNIVERSITY DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
% INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY
% PARTICULAR PURPOSE.  IN NO EVENT SHALL THE AUTHOR OR BROWN UNIVERSITY BE LIABLE FOR
% ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
% WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
% ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
% OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.  


% this.alpha           = 0;

  % Frame size
  sz = [size(this.images, 1), size(this.images, 2)];

  % If we have no initialization argument, initialize with all zeros
  if (nargin < 2)
    uv = zeros([sz, 2]);
  else
    uv = init;
  end
  
  layer_uv = cell(this.nLayers, 1);  
  for i = 1:this.nLayers;
      layer_uv{i} = uv;      
  end;  
  
  layerLabels = ones([sz this.nLayers]); % random initialization
  layerLabels = layerLabels./repmat(sum(layerLabels,3), [1 1 this.nLayers]);
  
  
  % Perform structure texture decomposition to get the texture component
    
  if this.texture == true;
      images  = structure_texture_decomposition_rof( this.images, 1/8, 100, this.alp); 
      fprintf('.....');
  else
      images  = scale_image(this.images, 0, 255);      
  end;
    
  % Automatic determine pyramid level
  this.pyramid_levels  =  1 + floor( log(min(size(images, 1), size(images,2))/16) / log(this.pyramid_spacing) );    

% Initialize segmentation from K-means flow vectors + intensity as features
this.pyramid_levels  = 2; 

u = uv(:,:,1); v = uv(:,:,2);
im1 = this.images(:,:,1);
data = [u(:) v(:) im1(:)/50];
[y_cluster centers] = kmeans(data, this.nLayers);
seg = reshape(y_cluster, sz);
% figure; imagesc(seg);

nL = this.nLayers;
maxL = 1-1e-3;
% minL = (1-maxL)/(1-nL);
minL = (1-maxL)/(nL-1);
layerLabels = ones([sz this.nLayers]) * minL;

[indr indc] = ind2sub(sz, reshape(1:prod(sz), sz) );
indx = sub2ind([sz this.nLayers], indr, indc, seg); 
layerLabels(indx) = maxL;

layer_g = label2g(layerLabels, this.lambda4);

  % For gnc stage 1    
  factor            = sqrt(2);  % sqrt(3) worse
  smooth_sigma      = sqrt(this.pyramid_spacing)/factor;   % or sqrt(3) recommended by Manuel Werlberger 
  f                 = fspecial('gaussian', 2*round(1.5*smooth_sigma) +1, smooth_sigma);        
  pyramid_images    = compute_image_pyramid(images, f, this.pyramid_levels, 1/this.pyramid_spacing);
  % For segmentation purpose
  org_pyramid_images = compute_image_pyramid(this.images, f, this.pyramid_levels, 1/this.pyramid_spacing);

  
  % For gnc stage 2 to gnc_iters  
  smooth_sigma      = sqrt(this.gnc_pyramid_spacing)/factor;
  f                 = fspecial('gaussian', 2*round(1.5*smooth_sigma) +1, smooth_sigma);        
  gnc_pyramid_images= compute_image_pyramid(images, f, this.gnc_pyramid_levels, 1/this.gnc_pyramid_spacing); 
    
  % For segmentation purpose
  org_gnc_pyramid_images = compute_image_pyramid(this.images, f, this.gnc_pyramid_levels, 1/this.gnc_pyramid_spacing);   

   
  tic
  
  for ignc = 1:this.gnc_iters     
  
      if this.display
          disp(['GNC stage: ', num2str(ignc)])
      end
          
      if ignc == 1
          pyramid_levels = this.pyramid_levels;
          pyramid_spacing = this.pyramid_spacing;
      else
          pyramid_levels = this.gnc_pyramid_levels;
          pyramid_spacing = this.gnc_pyramid_spacing;
      end;
      
      % Iterate through all pyramid levels starting at the top
      for l = pyramid_levels:-1:1

          if this.display
              disp(['-Pyramid level: ', num2str(l)])
          end

          % Generate copy of algorithm with single pyramid level and the
          % appropriate subsampling
          small = this;
          
          
          if ignc == 1
              nsz               = [size(pyramid_images{l}, 1) size(pyramid_images{l}, 2)];
              small.images      = pyramid_images{l};                            
              small.max_linear  = 1;             % number of linearization performed per warping, 1 OK for quadratic formulation
              im1   = org_pyramid_images{l}(:,:,1);             

          else
              small.images         = gnc_pyramid_images{l};
              nsz   = [size(gnc_pyramid_images{l}, 1) size(gnc_pyramid_images{l}, 2)];
              im1   = org_gnc_pyramid_images{l}(:,:,1);

          end;

          small.seg = im1;                             

          % Rescale the flow field and labels
          for i1 = 1:this.nLayers;
              layer_uv{i1} = resample_flow(layer_uv{i1}, nsz);                                         
          end;
          layerLabels = imresize(layerLabels, nsz);                           
          % Normalize
          layerLabels = layerLabels./repmat(sum(layerLabels,3), [1 1 this.nLayers]);
          layer_g = imresize(layer_g, nsz);     
          
          
          % Segment static image
          [tmp small.seg] = msseg(im1,7,6.5, round(50/pyramid_spacing.^(l-1)) );
          
          %figure; imagesc(tmp);
          %figure; imagesc(small.seg);
          
                    
          for i = 1:this.max_iters
              
            % Re-estimate layer label (ownership)                      
            if l ~= pyramid_levels | i ~=1                                
                %layerLabels = estimate_layer_labels_basic_gd(small, layer_uv, layerLabels);                
                layer_g = estimate_layer_labels_multi_layers_gd(small, layer_uv, layer_g);                                                
                layerLabels = g2labels(layer_g, this.lambda4);
                
                
                % for debug
%                 for i = 1:nL
%                     tmp = layerLabels(:,:,i);
%                     fprintf('%3.3e\t %3.3e\t', min(tmp(:)), max(tmp(:)));
%                 end;
%                 fprintf('\n');

            end;                             
           
            for i1 = 1:this.nLayers;  % last layer is occlusion layer
                
                uv = layer_uv{i1};
                small.layerLabel = layerLabels(:,:,i1);
                small.rho_data  = robust_function('quadratic', this.sigmaNs(i1));               
              
                % Run flow method on subsampled images
                uv = compute_flow_base(small, uv);   
                
                uv = real(uv); % for temporary debugging
                
                layer_uv{i1} = uv;
                
            end;               

            % for debug output
%             for i1 = 1:this.nLayers;
%                 figure; imshow(flowToColor(layer_uv{i1}), []);
%             end;
%             [tmp ind ] = max(layerLabels, [], 3);
%             figure; imagesc(ind);
%             if l == 1
%                 
%                 for i1 = 1:this.nLayers;
%                     
%                     if nargin == 3 & l == 1
%                         uv = layer_uv{i1};
%                         [aae stdae aepe] = flowAngErr(gt(:,:,1), gt(:,:,2), uv(:,:,1), uv(:,:,2), 0);        % ignore 0 boundary pixels
%                         fprintf('AAE %3.3f STD %3.3f average end point error %3.3f \n', aae, stdae, aepe);
%                     end;
%                     
%                 end;
%             end
            
          end

         
          if l == 1
              
              for i1 = 1:this.nLayers;
%                   figure; imshow(flowToColor(layer_uv{i1}), []);
%                   It = partial_deriv(small.images, layer_uv{i1}, this.interpolation_method, this.deriv_filter);
%                   figure; imshow(abs(It), []);
%                 fn = sprintf('layer_uv_%04d.mat', round(rand(1,1)*10000));
%                 save(fn, 'layer_uv');
                  
                  % figure; imshow(layerLabels(:,:,i1), []);
                  if nargin == 3 & l == 1
                      uv = layer_uv{i1};
                      [aae stdae aepe] = flowAngErr(gt(:,:,1), gt(:,:,2), uv(:,:,1), uv(:,:,2), 0);        % ignore 0 boundary pixels
                      fprintf('AAE %3.3f STD %3.3f average end point error %3.3f \n', aae, stdae, aepe);
                  end;

              end;
          end;
      
      
      end;
      
      
      % Compose final flow field
      
      % Assign occlusion layer   
      [tmp ind ] = max(layerLabels, [], 3);
      %figure; imagesc(ind);
      seg(:,:,1) = ind;
%       ind = assign_occlusion_layer(small, ind);
%       %figure; imagesc(ind);
%       seg(:,:,2) = ind;
      
%      if nargout == 2
          %seg = ind;
%      end;

      ind = repmat(ind, [1 1 2]);
      uv  = layer_uv{1};
      for i1 = 2:this.nLayers;
          uv(ind ==i1) = layer_uv{i1}(ind==i1);
      end;
      
      % Update GNC parameters (linearly)
      if this.gnc_iters > 1
          new_alpha  = 1 - ignc / (this.gnc_iters-1);
          this.alpha = min(this.alpha, new_alpha);
          this.alpha = max(0, this.alpha);          
      end;
      
      if true; %this.display

          fprintf('log-posterior\t%3.3e\n', evaluate_log_posterior(small, uv));
          if nargin == 3
              [aae stdae aepe] = flowAngErr(gt(:,:,1), gt(:,:,2), uv(:,:,1), uv(:,:,2), 0);        % ignore 0 boundary pixels
              fprintf('AAE %3.3f STD %3.3f average end point error %3.3f \n', aae, stdae, aepe);
          end;
          
          %figure; imshow(flowToColor(uv), []);                    
              
      end;

      toc/60

  end;  