function uv = compute_flow(this, init, gt)
%
%COMPUTE_FLOW   Compute flow field
%   UV = COMPUTE_FLOW(THIS[, INIT]) computes the flow field UV with
%   algorithm THIS and the optional initialization INIT.
%  
%   This is a member function of the class 'ba_optical_flow'. 

%   Author: Deqing Sun, Department of Computer Science, Brown University
%   Contact: dqsun@cs.brown.edu
%   $Date: 2007-- $

% Copyright 2007-2008, Brown University, Providence, RI.
%
%                         All Rights Reserved
%
% Permission to use, copy, modify, and distribute this software and its
% documentation for any purpose other than its incorporation into a
% commercial product is hereby granted without fee, provided that the
% above copyright notice appear in all copies and that both that
% copyright notice and this permission notice appear in supporting
% documentation, and that the name of the author and Brown University not be used in
% advertising or publicity pertaining to distribution of the software
% without specific, written prior permission.
%
% THE AUTHOR AND BROWN UNIVERSITY DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
% INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY
% PARTICULAR PURPOSE.  IN NO EVENT SHALL THE AUTHOR OR BROWN UNIVERSITY BE LIABLE FOR
% ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
% WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
% ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
% OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.  


  % Frame size
  sz = [size(this.images, 1), size(this.images, 2)];

  % If we have no initialization argument, initialize with all zeros
  if (nargin < 2)
    uv = zeros([sz, 2]);
  else
    uv = init;
  end
   
  % Perform structure texture decomposition to get the texture component
  if this.texture == true;
%       images  = structure_texture_decomposition_rof( this.images); 
      
      % 1 fixed filter response constancy
%       f = fspecial('gaussian', [5 5], 1.5);
      Dx = [-1 0 1]; Dy = Dx';
      f = Dx;
      images  = this.images- imfilter(this.images, f, 'symmetric');
      images  = scale_image(images, 0, 255);
      
      f = Dy;
      images2  = this.images- imfilter(this.images, f, 'symmetric');
      images2  = scale_image(images2, 0, 255);
      
  else
      images  = scale_image(this.images, 0, 255);      
  end;
    
  % Automatic determine pyramid level
  this.pyramid_levels  =  1 + floor( log(min(size(images, 1), size(images,2))/16) / log(this.pyramid_spacing) );  
  
  
% % To use initial value
% if sum(init(:) ~=0) > 0
%   this.pyramid_levels = 1;
%   this.alpha = 0;
%   this.gnc_iters = 1;
% end;

  % Construct image pyramid, using filter setting in Bruhn et al in "Lucas/Kanade.." (IJCV2005') page 218
  
  % For gnc stage 1    
  factor            = sqrt(2);  % sqrt(3)
  smooth_sigma      = sqrt(this.pyramid_spacing)/factor;   % or sqrt(3) recommended by Manuel Werlberger 
  f                 = fspecial('gaussian', 2*round(1.5*smooth_sigma) +1, smooth_sigma);        
  pyramid_images    = compute_image_pyramid(images, f, this.pyramid_levels, 1/this.pyramid_spacing);

  pyramid_images2    = compute_image_pyramid(images2, f, this.pyramid_levels, 1/this.pyramid_spacing);
  
  % For segmentation purpose
  org_pyramid_images = compute_image_pyramid(this.images, f, this.pyramid_levels, 1/this.pyramid_spacing);
  
  % For gnc stage 2 to gnc_iters  
  smooth_sigma      = sqrt(this.gnc_pyramid_spacing)/factor;
  f                 = fspecial('gaussian', 2*round(1.5*smooth_sigma) +1, smooth_sigma);        
  gnc_pyramid_images= compute_image_pyramid(images, f, this.gnc_pyramid_levels, 1/this.gnc_pyramid_spacing); 

  gnc_pyramid_images2= compute_image_pyramid(images2, f, this.gnc_pyramid_levels, 1/this.gnc_pyramid_spacing); 
  
  % For segmentation purpose
  org_gnc_pyramid_images = compute_image_pyramid(this.images, f, this.gnc_pyramid_levels, 1/this.gnc_pyramid_spacing);   
 
  tic
  
  uv2 = uv;
  
  for ignc = 1:this.gnc_iters     
  
      if this.display
          disp(['GNC stage: ', num2str(ignc)])
      end
          
      if ignc == 1
          pyramid_levels = this.pyramid_levels;
          pyramid_spacing = this.pyramid_spacing;
      else
          pyramid_levels = this.gnc_pyramid_levels;
          pyramid_spacing = this.gnc_pyramid_spacing;
      end;
      
%   if ignc == this.gnc_iters     % if 2nd GNC produces good initialization
%       pyramid_levels = 1;
%   end;
      
      % Iterate through all pyramid levels starting at the top
      for l = pyramid_levels:-1:1

          if this.display
              disp(['-Pyramid level: ', num2str(l)])
          end

          % Generate copy of algorithm with single pyramid level and the
          % appropriate subsampling
          small = this;
          small2 = this;

          
          if ignc == 1
              nsz               = [size(pyramid_images{l}, 1) size(pyramid_images{l}, 2)];
              small.images      = pyramid_images{l};                            
              small.max_linear  = 1;             % number of linearization performed per warping, 1 OK for quadratic formulation
              im1   = org_pyramid_images{l}(:,:,1);
              
              small2.images      = pyramid_images2{l};                            
          else
              small.images         = gnc_pyramid_images{l};
              nsz   = [size(gnc_pyramid_images{l}, 1) size(gnc_pyramid_images{l}, 2)];
              im1   = org_gnc_pyramid_images{l}(:,:,1);
              
              small2.images         = gnc_pyramid_images2{l};
          end;
          
          % Rescale the flow field
          uv        = resample_flow(uv, nsz);
          uv2 = uv;
          %uv2       = resample_flow(uv2, nsz);
              
%           [tmp small.seg] = msseg(im1,7,6.5, round(50/pyramid_spacing.^(l-1)) );
%           figure; imagesc(small.seg);
%           small.seg = im1;
          small.seg = small.images(:,:,1);
      
          
% % for debug area term
% if ignc < this.gnc_iters
%     small.median_filter_size = this.median_filter_size;
% else
%     small.median_filter_size = [];
% end;
        
          % Run flow method on subsampled images
          uv = compute_flow_base(small, uv);
          
          uv2 = compute_flow_base(small2, uv2);
          
          % Dy constancy computes u more accurately
          uv(:,:,1) = uv2(:,:,1);
      end
      
      
      % Update GNC parameters (linearly)
      if this.gnc_iters > 1
          new_alpha  = 1 - ignc / (this.gnc_iters-1);
          this.alpha = min(this.alpha, new_alpha);
          this.alpha = max(0, this.alpha);          
%             alphas = [1-1e-3 0 0];
%             this.alpha = alphas(ignc)
      end;

      if true; %this.display

          fprintf('log-posterior\t%3.3e\n', evaluate_log_posterior(small, uv));
          if nargin == 3
              [aae stdae aepe] = flowAngErr(gt(:,:,1), gt(:,:,2), uv(:,:,1), uv(:,:,2), 0);        % ignore 0 boundary pixels
              fprintf('AAE %3.3f STD %3.3f average end point error %3.3f \n', aae, stdae, aepe);
          end;
          
          figure; imshow(flowToColor(uv), []);          
          
%           if sum(init(:) ~=0) > 0
%               fprintf('Initial log-posterior\t%3.3e\n', evaluate_log_posterior(small, init));
%               if nargin == 3
%                   [aae stdae aepe] = flowAngErr(gt(:,:,1), gt(:,:,2), init(:,:,1), init(:,:,2), 0);        % ignore 0 boundary pixels
%                   fprintf('AAE %3.3f STD %3.3f average end point error %3.3f \n', aae, stdae, aepe);
%               end;
%           end;
              
      end;

      toc/60
  end;
  
  
  
  % Perform median filtering to remove outliers on the final result
  if ~isempty(this.median_filter_size)
      
%       for mf = 1:5;
                 
      uv(:,:,1) = medfilt2(uv(:,:,1), this.median_filter_size, 'symmetric');
      uv(:,:,2) = medfilt2(uv(:,:,2), this.median_filter_size, 'symmetric');

%       uv = robust_affine_fit(uv, this.median_filter_size);
      
% uv = denoise_medfilt2(uv, this.median_filter_size, 1);              
      
% uv = denoise_lorentzian(uv, this.median_filter_size, 10);       
%       end;
        % perform median filtering using intensity/segmentation information
%         uv = intensity_medfilt2(uv, this.imfsz, this.images(:,:,1), this.mfT);
%         uv = intensity_medfilt2(uv, this.median_filter_size, small.seg, this.mfT);
        
%   else
%       uv(:,:,1) = medfilt2(uv(:,:,1), [3 3], 'symmetric');
%       uv(:,:,2) = medfilt2(uv(:,:,2), [3 3], 'symmetric');
    % No extra iteration
%   elseif this.denoise_robust
%       uv(:,:,1) = denoise_robust(uv(:,:,1), this.lambda_dr);
%       uv(:,:,2) = denoise_robust(uv(:,:,2), this.lambda_dr);

  end;
    