function layerLabels = estimate_layer_labels(this, layer_uv, prev_layerLabels)
%%

% Estimate the expected layer labels

niters = 30;
delta  = 0.015;

% Compute BCE for each layer

for i = 1:this.nLayers
    It = partial_deriv(this.images, layer_uv{i}, this.interpolation_method, this.deriv_filter);
    
    % Simple version: no spatial interaction
    a = this.sigmaNs(i);
    layerLabels(:,:,i) = exp(- It.^2/2/a^2) / (sqrt(2*pi)*a) * this.pis(i);
    aIt(:,:,i)         = It;
end;

% Normalize
layerLabels = layerLabels./repmat(sum(layerLabels,3), [1 1 this.nLayers]);

% Blur for some spatial coherence

% Avoid numerical issue
minL = 1e-6;
layerLabels(layerLabels< minL) = minL;
% layerLabels(layerLabels>1)  = 1;       
% Normalize
layerLabels = layerLabels./repmat(sum(layerLabels,3), [1 1 this.nLayers]);

ec = compute_free_energy(layerLabels, this.sigmaNs, aIt);
ep = compute_free_energy(prev_layerLabels, this.sigmaNs, aIt);

if ec > ep
    % current solution has higher energy than previous
    layerLabels = prev_layerLabels;
end;

for iter = 1:niters
    [e g] = compute_free_energy(layerLabels, this.sigmaNs, aIt);
    layerLabels = layerLabels - delta * g;
    
    % Clip to [minL 1]
    layerLabels(layerLabels<minL)  = minL;
    layerLabels(layerLabels>1)  = 1;
    
end;


% Normalize
layerLabels = layerLabels./repmat(sum(layerLabels,3), [1 1 this.nLayers]);




function [e g] = compute_free_energy(layerLabels, sigmaNs, aIt)

% Compute the free energy defined in Y. Weiss and E.H. Adelson CVPR 96

lambda = 1;     % weight on spatial term
lambda2 = 1;    % weight on sum-to-one term


% Reshape labels
layerLabels = reshape(layerLabels, size(aIt));

nL = size(aIt, 3);

e = 0;
% Compute energy

% Data term
for i = 1:nL
    e = e+ sum(sum(aIt(:,:,i).^2.*layerLabels(:,:, i))) / (2*sigmaNs(i)^2);
end;

% coherence with vertical neighbors
tmp = layerLabels(1:end-1, :, :).*layerLabels(2:end, :, :);
e = e - lambda * sum(tmp(:));
% % coherence with vertical neighbors
tmp = layerLabels(:, 1:end-1, :).*layerLabels(:, 2:end, :);
e = e - lambda * sum(tmp(:));

% entropy
e = e + sum(layerLabels(:).*log(layerLabels(:)));

% singleton 
% e = e + sum(1./layerLabels(:))/4;   % L1

% sum-to-one term
tmp = sum(layerLabels, 3);
e   = e + lambda2 * sum((tmp(:)-1).^2);

% Compute gradient w.r.t. layers
if nargout > 1 
    g = zeros(size(aIt));
    
    % Spatial coherence
    g(1:end-1, :,:) = g(1:end-1, :,:) -layerLabels(2:end, :,:);   % top  neighbor
    g(2:end, :,:)   = g(2:end, :,:) -layerLabels(1:end-1, :,:); % bottom neighbor
    
    g(:, 1:end-1, :) = g(:, 1:end-1, :)-layerLabels(:, 2:end, :);  % right neighbor
    g(:, 2:end, :)   = g(:, 2:end, :)  -layerLabels(:, 1:end-1,:); % left neighbor
    
    g = g* lambda ;
    
    % Data term
    for i = 1:nL
        g(:,:,i) = g(:,:,i) + aIt(:,:,i).^2 / (2*sigmaNs(i)^2);
    end;         
     
    % Entropy
    g = g + log(layerLabels) + 1;

    % singleton
    % g = g - 1./layerLabels.^2/4;

    % sum to one
    g = g + 2 * lambda2 * repmat(sum(layerLabels, 3) -1, [ 1 1 nL]);

    %g = g(:);
end;
    

