function [layer_q MFOBJ] = perform_mean_field_inference_2layers_prealloc_FB...
    (mfParams, pyramid_images,...
    color_pyramid_images, layer_uv, layer_uvb, ini_layer_g, ini_layer_q)
%%
% perform mean field approximation to infer the approximate marginal
% probabilities for the layer support function given the flow field

% for the special two-layer case

% forward and backward
% odd iteration  t = 1 : T
% even iteration t = T: -1 : 1

% convert initial value to confidence

H  = size(layer_uv{1},1);
W  = size(layer_uv{1},2);
T  = length(layer_uv) + 1;
nL = size(layer_uv{1},4);

[X Y] = meshgrid(1:W, 1:H);

nValues = 2; % only 0 and 1 for Ising model
if nL~=2;
    fprintf('perform_mean_field_..: function tailored to 2 layers!');
end;

colorImages = color_pyramid_images{1};
[dataCost warpedDataCost BLMtx oob...
    dataCostB BLMtxB warpedDataCostB oobB] =...
    precompute_mean_field(mfParams, layer_uv, layer_uvb,...
    pyramid_images, T, nL);

if nargin > 6 && ~isempty(ini_layer_q)
    layer_q = ini_layer_q;
elseif nargin > 5
    layer_q = ones([H W nL-1 T nValues])*1e-3;
    layer_q(:,:,:,:,1) = ini_layer_g ==1;
    layer_q(:,:,:,:,2) = ini_layer_g ==0;
    layer_q = bsxfun(@rdivide, layer_q, sum(layer_q,5)+mfParams.minValueQ);
else    
    layer_q = ones([H W nL-1 T nValues])*0.5;
    for t=1:T-1        
        layer_q(:,:,1,t,1) = dataCost{t}{1} <= dataCost{t}{2};
        layer_q(:,:,1,t,2) = dataCost{t}{1} > dataCost{t}{2};
    end;
    t = T;
    layer_q(:,:,1,t,1) = dataCostB{t-1}{1} < dataCostB{t-1}{2};
    layer_q(:,:,1,t,2) = dataCostB{t-1}{1} > dataCostB{t-1}{2};
    layer_q = bsxfun(@rdivide, layer_q, sum(layer_q,5)+mfParams.minValueQ);
end;

% for debug, show initial value
% figure; imshow([layer_q(:,:,1,1,1) layer_q(:,:,1,1, 2)], []);

curr_layer_q = layer_q;

% Construct feature array for preallocated mean-field.
features_mf = {};
for t=1:T
    % used instead of features
    features_mf{t} = cat(3, squeeze(colorImages(:,:,:,t))/mfParams.sigma_i, ...
        X/mfParams.sigma_s, Y/mfParams.sigma_s);
end
% used instead of features2
features_mf{T+1} = cat(3, X/mfParams.sigma_spa, Y/mfParams.sigma_spa);

% Create parameter array
mf_prealloc_params = [repmat(mfParams.lambdaSpa, 1, T) mfParams.lambdaSpaSeg];

% Pre-allocate mean field.
perform_dense_message_passing_mex_prealloc(1, 'a', features_mf, mf_prealloc_params);

% Pre-allocate mean field for evaluation
perform_dense_message_passing_mex_prealloc(2, 'a', ...
    features_mf(1:end-1), mf_prealloc_params(1:end-1));

lambdaTemps = exp(floor(linspace(log10(mfParams.lambdaTempStart),...
    log10(mfParams.lambdaTempEnd)+1-1e-3, mfParams.mfinnerIters))*log(10));
expScales   =  logspace(log10(mfParams.expScaleStart),...
    log10(mfParams.expScaleEnd),  mfParams.mfinnerIters);

if nargout > 1
    % use same lambda for evaluating lower bound (k-l divergence)
    mfParams.lambdaTemp = mfParams.lambdaTempEnd;
    
    mfObj = evaluate_mean_field_objective_2layers_prealloc(layer_q, mfParams,...
        dataCost,BLMtx, dataCostB, BLMtxB, colorImages, X, Y);
    
    MFOBJ(1) = mfObj;

end;
iterloop = tic;

for iter = 1:mfParams.mfinnerIters

    % annealing
    mfParams.lambdaTemp = lambdaTemps(iter);
    mfParams.expScale   = expScales(iter);

    % spatial message passing, fully connected prior    
    tmp_out = zeros(H,W, nValues);
    
    if mod(iter, 2)==1
        sequence = 1:T; % forward in time
    else
        sequence = T:-1:1; % backward in time
    end;

    for iQ = 1:length(sequence)
        t = sequence(iQ);       
        
        curr_q = layer_q(:,:,:,t,:);
        
        for layerID=1:nL-1
            tmp = squeeze(layer_q(:,:,layerID,t,:));
            
            perform_dense_message_passing_mex_prealloc( ...
                1, 'p', t, tmp, tmp_out);
            
            if mfParams.spaSubtract
                % message passing needs to subtract its own * weight
                curr_q(:,:,layerID,:,:) = tmp_out(:,:,end:-1:1)-...
                    mfParams.lambdaSpa*tmp(:,:,end:-1:1);
            else
                % not easy to subtract the exact value due to weight
                curr_q(:,:,layerID,:,:) = tmp_out(:,:,end:-1:1);
            end;
            
            perform_dense_message_passing_mex_prealloc( ...
                1, 'p', T+1, tmp, tmp_out);
            
            tmp_out = squeeze(curr_q(:,:,layerID,:,end:-1:1))+ tmp_out;
            
            if mfParams.spaSubtract
                % message passing needs to subtract its own * weight
                curr_layer_q(:,:,layerID,t,:) = tmp_out(:,:,end:-1:1)-...
                    mfParams.lambdaSpa*tmp(:,:,end:-1:1);
            else
                % not easy to subtract the exact value due to weight
                curr_q(:,:,layerID,:,:)  =...
                    tmp_out(:,:,end:-1:1);
            end;
            
        end;
        
        %%%%%%% forward flow field
        curr_q = curr_q + pass_message_flow_one_frame(mfParams, layer_q,...
        dataCost, BLMtx, warpedDataCost, t, T, nValues, curr_q);
        
       
        if mfParams.useBackFlow
            %%%%%%% backward flow field
            curr_q = curr_q + pass_message_flow_backward_one_frame(...
                 mfParams, layer_q, dataCostB, BLMtxB, warpedDataCostB,...
                 t, T, nValues, curr_q);
        end;               

        curr_q = curr_q -min(curr_q(:));
        curr_q = exp(-curr_q*mfParams.expScale);
        
        % avoid divide by zero -> NaN
        curr_q(curr_q<mfParams.minValueQ) = mfParams.minValueQ;     
        curr_q = bsxfun(@rdivide, curr_q, sum(curr_q,5));
        
        if mfParams.mixStepSize < 1            
            curr_q = mfParams.mixStepSize*curr_q + ...
                (1-mfParams.mixStepSize)*layer_q(:,:,1,t,:);
        end;
        
        if mfParams.medflt && (iter < mfParams.mfinnerIters) ...
                && (lambdaTemps(iter)~=lambdaTemps(iter+1))
            for qID = 1:nValues
                curr_q(:,:,:,:,qID) = medfilt2(...
                    curr_q(:,:,:,:,qID), mfParams.mfsize, 'symmetric');
            end;
        end    

        % avoid divide by zero -> NaN
        curr_q(curr_q<mfParams.minValueQ) = mfParams.minValueQ;
        curr_q = bsxfun(@rdivide, curr_q, sum(curr_q,5));
        
        layer_q(:,:,:,t,:) = curr_q;
    end;
    
    if nargout > 1
        % use same lambda for evaluating lower bound (k-l divergence)
        mfParams.lambdaTemp = mfParams.lambdaTempEnd;
        
        mfObj = evaluate_mean_field_objective_2layers_prealloc(layer_q,...
            mfParams, dataCost,BLMtx, dataCostB, BLMtxB, colorImages, X, Y);        
       
        MFOBJ(iter+1) = mfObj;
        
        if iter >1 && abs(mfObj-MFOBJ(iter-1))/abs(mfObj) < 1e-6
            break;
        end;
    end;

end;

% fprintf('Time elapsed in iteration loop: %d seconds\n', toc(iterloop));

% Dealloc mean field
perform_dense_message_passing_mex_prealloc(1, 'd');
perform_dense_message_passing_mex_prealloc(2,'d');